/* ictok.c */
/***********************************************
* copy next C token between boundaries into wbuf
* Istvan Mohos, 1987 --- in the Public Domain
***********************************************/

#include "i.h"
#include "icmapalnum.h"

char *
ictok (start, end, wbuf)
register char *start;
register char *end;
char *wbuf;
{
	char *icalntok();
	register char *rest;
	register char *wp = wbuf;
	int runaway;

	if (NULCHARP (start))
		return(NULL);
	ITOEND;
	if (start == end)
		return(NULL);

	if ((rest = icalntok (start, end, wbuf)) == NULL)
		return (NULL);
	if (rest == end)
		return (rest);

	switch (*wbuf) {
		/* normal word */
		default: return (rest);

		/* if the current token is a "normal" word beginning with a
		   digit, look ahead to complete a floating point number
		*/
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			switch (*rest) {
				default: return (rest);
				case '.':
					wp = inull (wbuf);
					*wp = *rest++;
					while (rest != end && *rest >= '0' && *rest <= '9')
						*++wp = *rest++;
					*++wp = '\0';
					return (rest);
			}

		/* similarly, complete a float if it begins with '.' */
		case '.':
			while (rest != end && *rest >= '0' && *rest <= '9')
				*++wp = *rest++;
			*++wp = '\0';
			return (rest);

		/* quote starts a token that ends with the next non-escaped
		   quote.  If the matching quote is not found by byte 511,
		   "runaway single|double quote" is substituted for the token,
		   and a pointer to the 512th byte is returned.
		*/
		case '\'':
			runaway = 1;
			escape1:
			while (rest != end && *rest != '\'') {
				*++wp = *rest++;
				if (++runaway == 511) {
					strcpy (wbuf, "runaway single quote");
					return (rest);
				}
			}
			if (rest != end) { /* sitting on next quote character */
				*++wp = *rest++;
				if (*(rest-2) == '\\') {
					if (++runaway == 511) {
						strcpy (wbuf, "runaway single quote");
						return (rest);
					}
					goto escape1;
				}
			}
			*++wp = '\0';
			return (rest);

		case '"':
			runaway = 1;
			escape2:
			while (rest != end && *rest != '"') {
				*++wp = *rest++;
				if (++runaway == 511) {
					strcpy (wbuf, "runaway double quote");
					return (rest);
				}
			}
			if (rest != end) { /* sitting on next quote character */
				*++wp = *rest++;
				if (*(rest-2) == '\\') {
					if (++runaway == 511) {
						strcpy (wbuf, "runaway double quote");
						return (rest);
					}
					goto escape2;
				}
			}
			*++wp = '\0';
			return (rest);

		/* CONSTANTS \<CR> \" \' \<octal> \\ \b \f \n \r \t \v
		   Rule: token is following byte concatenated to backslash
				 except when following byte is an octal digit
		*/
		case '\\':
			if (*rest >= '0' && *rest <= '7') {
				while (rest != end && *rest >= '0' && *rest <= '7')
					*++wp = *rest++;
				*++wp = '\0';
				return (rest);
			}
			else {
				*++wp = *rest;
				*++wp = '\0';
				return (++rest);
			}

		/* OPERATORS != %= && &= *= /= ^= ++ += -- -= -> |= || ==
					 << <<= <= >= >> >>=
		   ARCHAIC   =* =+ =-
		   COMMENT DELIMITERS (delimiting these lines of text)
		*/
		case '!':
		case '%':
			switch (*rest) {
				default: return (rest);
				case '=':
					*++wp = *rest;
					*++wp = '\0';
					return (++rest);
			}
		case '&':
			switch (*rest) {
				default: return (rest);
				case '&':
				case '=':
					*++wp = *rest;
					*++wp = '\0';
					return (++rest);
			}
		case '*':
			switch (*rest) {
				default: return (rest);
				case '=':
				case '/':
					*++wp = *rest;
					*++wp = '\0';
					return (++rest);
			}
		case '/':
			switch (*rest) {
				default: return (rest);
				case '=':
				case '*':
					*++wp = *rest;
					*++wp = '\0';
					return (++rest);
			}
		case '^':
			switch (*rest) {
				default: return (rest);
				case '=':
					*++wp = *rest;
					*++wp = '\0';
					return (++rest);
			}
		case '+':
			switch (*rest) {
				default: return (rest);
				case '+':
				case '=':
					*++wp = *rest;
					*++wp = '\0';
					return (++rest);
			}
		case '-':
			switch (*rest) {
				default: return (rest);
				case '-':
				case '=':
				case '>':
					*++wp = *rest;
					*++wp = '\0';
					return (++rest);
			}
		case '|':
			switch (*rest) {
				default: return (rest);
				case '|':
				case '=':
					*++wp = *rest;
					*++wp = '\0';
					return (++rest);
			}
		case '=':
			switch (*rest) {
				default: return (rest);
				case '*':
				case '+':
				case '-':
				case '=':
					*++wp = *rest;
					*++wp = '\0';
					return (++rest);
			}
		case '<':
			switch (*rest) {
				default: return (rest);
				case '=':
					*++wp = *rest;
					*++wp = '\0';
					return (++rest);
				case '<':
					*++wp = *rest++;
					if (rest == end) {
						*++wp = '\0';
						return (rest);
					}
					if (*rest == '=')
						*++wp = *rest++;
					*++wp = '\0';
					return (rest);
			}
		case '>':
			switch (*rest) {
				default: return (rest);
				case '=':
					*++wp = *rest;
					*++wp = '\0';
					return (++rest);
				case '>':
					*++wp = *rest++;
					if (rest == end) {
						*++wp = '\0';
						return (rest);
					}
					if (*rest == '=')
						*++wp = *rest++;
					*++wp = '\0';
					return (rest);
			}
	}
}

static
char *
icalntok (start, end, wbuf)
register char *start;
register char *end;
char *wbuf;
{
	register char *rp;
	register char *anchor;

	if (start >= end)
		return(NULL);

	/* stop at first non-white character, drop anchor */
	for (;WHITE(*start);)
		if (++start == end)
			return(NULL);
	anchor = start;

	/* stop at first non-alnum, white, or at end */
	for (rp = cmapalnum; *(rp + *start);)
		if (++start == end)
			break;

	/* get a single_byte, non-alnum token; anchor and start can
	   only point to the same address if start found a
	   single-byte non-alnum token in the previous loop
	*/
	if (anchor==start)
		++start;

	/* reuse pointers, copy token into wbuf */
	end = start;
	rp = wbuf;
	for (start = anchor; start < end;)
		*rp++ = *start++;
	*rp = '\0';

	return(start);
}
