/* idamage.c */
/**************************************************************
* return TRUE if passed error number means probable file damage
* Istvan Mohos, 1989 --- in the Public Domain
**************************************************************/

#include "i.h"
#include <errno.h>

int
idamage (error)
int error;
{
	/* failed during read, malloc, etc., no file damage */
	switch (error) {
	case -EPERM:        /* not owner */
	case -ENOENT:       /* no such file */
	case -EACCES:       /* permission denied; note one 'S' at end */
	case -EISDIR:       /* attempt to write on directory */
	case -ENFILE:       /* file table overflow */
	case -EMFILE:       /* file descriptor overflow */
	case -ETXTBSY:      /* attempt to write on executing obj */
	case -EROFS:        /* read only file system */
#ifndef REALUNIX
	case -ENAMETOOLONG: /* BSD component 255, 1023 total path */
#endif
		return (0);

	/* probable file damage */
	case -EINTR:        /* interrupt (while writing) */
	case -EIO:          /* I/O error */
	case -EFBIG:        /* file too big > 1082,201,088 */
	case -ENOSPC:       /* no space on disk */
#ifndef REALUNIX
	case -EDQUOT:       /* user disk quota exhausted */
#endif
		return (1);

	/* <= -sys_nerr: not system call: null file name, empty file
	   all non-negative values: no error
	*/
	default:
		return (0);
	}
}
